"""
These tests check for possible problems with apps that use this package.

Usage
-----
You can import all of the tests in this module, and run them with the built in
test runner. E.g.

.. code-block:: python

   # my_app_tests.py
   import unittest

   from my_app.factory import create_web_app
   from arxiv.base.app_tests import *

   app = create_web_app()

   if __name__ == '__main__':
       with app.app_context():
           unittest.main()


And run with:

.. code-block:: bash

   pipenv run python my_app_tests.py


"""

from flask import Flask, current_app
from unittest import TestCase
from . import config


class TestConfig(TestCase):
    """Look for common problems with the app config."""

    def test_server_name(self):
        """SERVER_NAME should be None (or unset)."""
        self.assertIsNone(current_app.config.get('SERVER_NAME'),
                          "SERVER_NAME should be None (or unset)")


class TestStaticFiles(TestCase):
    """Make sure that static files are configured safely on an app."""

    def test_static_url_paths(self) -> None:
        """Check for possible issues with the static URL path."""
        self.assertTrue(current_app.static_url_path.startswith('/static'),
                        "The static url path for the app should start with"
                        " /static")
        self.assertIn(
            f'/{current_app.name}', current_app.static_url_path,
            f"The static_url_path should include the name of the app"
        )
        self.assertIn(
            f'/{current_app.config.get("APP_VERSION")}',
            current_app.static_url_path,
            f"The static_url_path should include the version of the app"
            " (config.APP_VERSION)"
        )

        for key, blueprint in current_app.blueprints.items():
            if key == 'base':
                self.assertEqual(
                    blueprint.static_url_path,
                    f"/static/base/{config.BASE_VERSION}",
                    "The static URL for base should be independent of the"
                    " app static URL path, and should be based on the"
                    " version of the arxiv.base package."
                )

            else:
                self.assertIsNotNone(
                    blueprint.static_url_path,
                    f"The static_url_path should be set on blueprint `{key}`"
                )
                self.assertTrue(
                    current_app.static_url_path.startswith('/static'),
                    "The static url path for the app should start with /static"
                )
                self.assertIn(
                    f'/{current_app.name}', blueprint.static_url_path,
                    f"The static_url_path should include the name of the app"
                )
                self.assertIn(
                    f'/{current_app.config.get("APP_VERSION")}',
                    blueprint.static_url_path,
                    f"The static_url_path should include the version of the"
                    " app (config.APP_VERSION)"
                )
                self.assertIn(
                    key, blueprint.static_url_path,
                    f"The static_url_path for blueprint `{key}` should include"
                    " the name of the blueprint"
                )



"""Provides JSON Schema validation tools."""

import json
import os
from functools import wraps
from typing import Callable, Tuple, Any
import jsonschema
from flask import request

ValidationError = jsonschema.exceptions.ValidationError

HTTP_200_OK = 200
HTTP_400_BAD_REQUEST = 400


def load(schema_path: str) -> Callable:
    """
    Load a JSON Schema from ``schema_path`` and generate a validator.

    Parameters
    ----------
    schema_path : str
        Location of the target schema.

    Returns
    -------
    callable
        A validator function; when called with a ``dict``, validates the data
        against the schema.

    """
    try:
        with open(schema_path) as f:
            schema = json.load(f)
    except json.decoder.JSONDecodeError as e:
        raise IOError('Could not load %s: %s' % (schema_path, e)) from e

    schema_base_path = os.path.dirname(os.path.realpath(schema_path))
    resolver = jsonschema.RefResolver(referrer=schema,
                                      base_uri="file://%s/" % schema_base_path)

    def validate(data: dict) -> None:
        """
        Validate ``data`` against the enclosed schema.

        Parameters
        ----------
        data : dict

        Raises
        ------
        :class:`.ValidationError`

        """
        jsonschema.validate(data, schema, resolver=resolver)
    return validate


def validate_request(schema_path: str) -> Callable:
    """
    Generate a route decorator that validates the request body.

    Parameters
    ----------
    schema_path : str
        Path (absolute, or relative to the execution path) to the JSON Schema
        document.

    Returns
    -------
    decorator
        Decorates a Flask route with request body validation against the
        specified JSON Schema.

    Examples
    --------
    Should be used with a Flask route, for example:

        from util import schema

        @blueprint.route('/foo/', methods=['POST'])
        @schema.validate_request('schema/resources/foo.json')
        def some_route() -> Tuple[dict, int, dict]:
            ...

    """
    validate = load(schema_path)

    def _decorator(func: Callable) -> Callable:
        @wraps(func)
        def _wrapper(*args: Any, **kwargs: Any) -> Tuple[dict, int, dict]:
            try:
                validate(request.get_json())
            except ValidationError as e:
                # A summary of the exception is on the first line of the repr.
                msg = str(e).split('\n')[0]
                return (
                    {'reason': 'Metadata validation failed: %s' % msg},
                    HTTP_400_BAD_REQUEST,
                    {}
                )
            response: Tuple[dict, int, dict] = func(*args, **kwargs)
            return response
        return _wrapper
    return _decorator



"""
Functions to create hashes for clickthrough links.

These are used to log to the web acccess logs when the DOI and
bookmarking links are clicked.

The hash is used to prevent malicious use of the click through
controller to create links that look like they are on arXiv but get
redirected someplace undesirable.
"""

from typing import Callable
import hashlib
from flask import url_for, current_app


def create_hash(secret: str, url: str) -> str:
    """Create a hash of the secret and url."""
    s = f'{secret}{url}'
    return str(hashlib.md5(s.encode()).hexdigest()[0:8])


def is_hash_valid(secret: str, url: str, ct_hash: str) -> bool:
    """Check that ct_hash was generated by create_hash for secret and url."""
    return ct_hash == create_hash(secret, url)


def clickthrough_url(url: str) -> str:
    """Create a URL to the clickthrough service with a valid hash."""
    secret = current_app.config.get('CLICKTHROUGH_SECRET', 'foo')
    ct_url: str = url_for('clickthrough', url=url, v=create_hash(secret, url))
    return ct_url



"""Context processors."""

from flask import Flask

from arxiv.base import alerts, config


def inject_get_alerts() -> dict:
    """Insert :func:`alerts.get_alerts` into the template context."""
    return dict(get_alerts=alerts.get_alerts)


def inject_get_hidden_alerts() -> dict:
    """Insert :func:`alerts.get_hidden_alerts` into the template context."""
    return dict(get_hidden_alerts=alerts.get_hidden_alerts)

def inject_a11y_url() -> dict:
    """Insert local config variable `A11Y_URL` into the template context."""
    return dict(A11Y_URL=config.A11Y_URL)

def register_context_processors(app: Flask) -> None:
    """
    Register base context processors on a Flask app.

    Parameters
    ----------
    app : :class:`Flask`

    """
    app.context_processor(inject_get_alerts)
    app.context_processor(inject_get_hidden_alerts)
    app.context_processor(inject_a11y_url)



"""Provides base classes for WSGI middlewares."""

from typing import Callable, Union, Iterable, TypeVar
from flask import Flask


class BaseMiddleware(object):
    """
    Base class for WSGI middlewares.

    Child classes should implement one or both of:

    - ``before(environ: dict, start_response: Callable) -> Tuple[dict, Callable]``
    - ``after(response: Iterable) -> Iterable``

    """

    def __init__(self, app: Union[Flask, Callable]) -> None:
        """
        Set the app factory that this middleware wraps.

        Parameters
        ----------
        app : :class:`.Flask` or callable
            The application wrapped by this middleware. This might be an inner
            middleware, or the original :class:`.Flask` app itself.

        """
        self.app = app

    def __call__(self, environ: dict, start_response: Callable) -> Iterable:
        """
        Handle a WSGI request.

        Parameters
        ----------
        environ : dict
            WSGI request environment.
        start_response : function
            Function used to begin the HTTP response. See
            https://www.python.org/dev/peps/pep-0333/#the-start-response-callable

        Returns
        -------
        iterable
            Iterable that generates the HTTP response. See
            https://www.python.org/dev/peps/pep-0333/#the-application-framework-side
        """
        if hasattr(self, 'before'):
            environ, start_response = self.before(environ, start_response)  # type: ignore
        response: Iterable = self.app(environ, start_response)

        if hasattr(self, 'after'):
            response = self.after(response)       # type: ignore
        return response

    @property
    def wsgi_app(self):    # type: ignore
        """
        Refer to the current instance of this class.

        This is here for consistency with the :class:`.Flask` interface.
        """
        return self



"""Flask configuration."""

from typing import Optional
import os
from urllib.parse import urlparse

SERVER_NAME = None

SECRET_KEY = os.environ.get('SECRET_KEY', 'yes123')
A11Y_URL = os.environ.get("ARXIV_ACCESSIBILITY_URL",
                          "mailto:web-accessibility@cornell.edu")

EXTERNAL_URL_SCHEME = os.environ.get('EXTERNAL_URL_SCHEME', 'https')
BASE_SERVER = os.environ.get('BASE_SERVER', 'arxiv.org')

URLS = [
    ("help", "/help", BASE_SERVER),
    ("help_identifier", "/help/arxiv_identifier", BASE_SERVER),
    ("help_trackback", "/help/trackback", BASE_SERVER),
    ("help_mathjax", "/help/mathjax", BASE_SERVER),
    ("help_social_bookmarking", "/help/social_bookmarking", BASE_SERVER),
    ("contact", "/help/contact", BASE_SERVER),
    ("search_box", "/search", BASE_SERVER),
    ("search_archive", "/search/<archive>", BASE_SERVER),
    ("search_advanced", "/search/advanced", BASE_SERVER),
    ("account", "/user", BASE_SERVER),
    ("login", "/login", BASE_SERVER),
    ("logout", "/logout", BASE_SERVER),
    ("home", "/", BASE_SERVER),
    ("ignore_me", "/IgnoreMe", BASE_SERVER),    # Anti-robot honneypot.
    ("pdf", "/pdf/<arxiv:paper_id>", BASE_SERVER),
    ("twitter", "/arxiv", "twitter.com"),
    ("blog", "/arxiv", "blogs.cornell.edu"),
    ("wiki", "/display/arxivpub/arXiv+Public+Wiki", "confluence.cornell.edu"),
    ("library", "/", "library.cornell.edu"),
    ("university", "/", "cornell.edu"),
    ("acknowledgment", "/x/ALlRF", "confluence.cornell.edu"),
    ("faq", "/help/faq", BASE_SERVER),
    ("subscribe", "/help/subscribe", BASE_SERVER),
    ("submit", "/submit", BASE_SERVER),
    ("about", "/about", BASE_SERVER),
    ("team", "/about/people/leadership_team", BASE_SERVER),
    ("privacy_policy", "/help/policies/privacy_policy", BASE_SERVER),
    ("abs", "/abs/<arxiv:paper_id>v<string:version>", BASE_SERVER),
    ("abs_by_id", "/abs/<arxiv:paper_id>", BASE_SERVER),
    ("clickthrough", "/ct", BASE_SERVER)
]
"""
URLs for external services, for use with :func:`flask.url_for`.

For details, see :mod:`arxiv.base.urls`.
"""

# In order to provide something close to the config_url behavior, this will
# look for ARXIV_{endpoint}_URL variables in the environ, and update `URLS`
# accordingly.
for key, value in os.environ.items():
    if key.startswith('ARXIV_') and key.endswith('_URL'):
        endpoint = "_".join(key.split('_')[1:-1]).lower()
        o = urlparse(value)
        if not o.netloc:    # Doesn't raise an exception.
            continue
        i: Optional[int]
        try:
            i = list(zip(*URLS))[0].index(endpoint)
        except ValueError:
            i = None
        if i is not None:
            URLS[i] = (endpoint, o.path, o.netloc)

ARXIV_BUSINESS_TZ = os.environ.get("ARXIV_BUSINESS_TZ", "US/Eastern")

BASE_VERSION = "0.15.2"
"""The version of the arxiv-base package."""

APP_VERSION = "0.15.2"
"""The version of the base test app."""

"""
Flask-S3 plugin settings.

See `<https://flask-s3.readthedocs.io/en/latest/>`_.
"""
FLASKS3_BUCKET_NAME = os.environ.get('FLASKS3_BUCKET_NAME', 'some_bucket')
# FLASKS3_CDN_DOMAIN = os.environ.get('FLASKS3_CDN_DOMAIN', 'static.arxiv.org')
FLASKS3_USE_HTTPS = os.environ.get('FLASKS3_USE_HTTPS', 1)
FLASKS3_FORCE_MIMETYPE = os.environ.get('FLASKS3_FORCE_MIMETYPE', 1)
FLASKS3_ACTIVE = bool(int(os.environ.get('FLASKS3_ACTIVE', 0)))

# AWS credentials.
AWS_ACCESS_KEY_ID = os.environ.get('AWS_ACCESS_KEY_ID', 'nope')
AWS_SECRET_ACCESS_KEY = os.environ.get('AWS_SECRET_ACCESS_KEY', 'nope')
AWS_REGION = os.environ.get('AWS_REGION', 'us-east-1')



from unittest import TestCase, mock
import sys
from contextlib import contextmanager
from io import StringIO

import logging as pyLogging
from arxiv.base import logging


class TestGetLogger(TestCase):
    """Test :func:`.logging.getLogger`."""

    def test_get_logger_no_app_nor_request(self):
        """There is no application nor request context."""
        stream = StringIO()

        logger = logging.getLogger('foologger', stream)
        self.assertIsInstance(logger, pyLogging.Logger,
                              "Should return a logging.Logger instance")

        logger.error('foo')
        captured_value = stream.getvalue()
        stream.close()
        self.assertIn('ERROR: "foo"', captured_value,
                      "Should log normally even if request is not present")

    @mock.patch(f'{logging.__name__}.request')
    def test_get_logger_with_request(self, mock_request):
        """The request context is available."""
        mock_request.environ = {'REQUEST_ID': 'foo-id-1234'}
        stream = StringIO()
        logger = logging.getLogger('foologger', stream)
        self.assertIsInstance(logger, pyLogging.Logger,
                              "Should return a logging.Logger instance")
        logger.error('foo')
        captured_value = stream.getvalue()
        stream.close()
        self.assertIn('foo-id-1234', captured_value,
                      "Should include request ID in log messages")

    @mock.patch(f'{logging.__name__}.request')
    @mock.patch('arxiv.base.logging.get_application_config')
    def test_config_sets_loglevel(self, mock_get_config, mock_request):
        """LOGLEVEL param in config controls log level."""
        mock_get_config.return_value = {'LOGLEVEL': 10}
        mock_request.environ = {'REQUEST_ID': 'foo-id-1234'}
        stream = StringIO()
        logger = logging.getLogger('foologger', stream)
        logger.debug('foo')
        captured_value = stream.getvalue()
        stream.close()
        self.assertIn('DEBUG', captured_value,
                      "Changing LOGLEVEL in the app config should change the"
                      " logger log level")

    @mock.patch(f'{logging.__name__}.request')
    def test_paper_id_is_set(self, mock_request):
        """``paperid`` is included in the log data."""
        mock_request.environ = {'REQUEST_ID': 'foo-id-1234'}
        stream = StringIO()
        logger = logging.getLogger('foologger', stream)
        logger.error('what', extra={'paperid': '1234'})
        captured_value = stream.getvalue()
        stream.close()
        self.assertIn('arxiv:1234', captured_value,
                      "Should include paper ID in log messages")

    @mock.patch(f'{logging.__name__}.request')
    def test_paper_id_is_not_set(self, mock_request):
        """``paperid`` is not included in the log data."""
        mock_request.environ = {'REQUEST_ID': 'foo-id-1234'}
        stream = StringIO()
        logger = logging.getLogger('foologger', stream)
        logger.error('what')
        captured_value = stream.getvalue()
        stream.close()
        self.assertIn('arxiv:null', captured_value,
                      "Paper ID should be null in log messages")



"""
CSRF protection for arXiv forms.

To protect a form, inherit from :class:`CSRFForm`. When the form is handled,
this class expects that:

- There is an active :class:`flask.Request`.
- The object ``flask.request.session`` exists, and has a ``str`` attribute
  ``nonce``. See :class:`arxiv.users.domain.Session` and
  :class:`arxiv.users.auth.Auth`.
- A parameter called ``CSRF_SECRET`` is set in the application configuration.

Here's an example:

.. code-block:: python


   from typing import Tuple
   from werkzeug import MultiDict
   from wtforms import StringField
   from http import HTTPStatus as status
   from arxiv.forms import csrf

   ResponseData = Tuple[dict, int, dict]


   class ProtectedForm(csrf.CSRFForm):
       '''A CSRF-protected form.'''

       something_sensitive = StringField('Something sensitive')


   def some_controller(method: str, form_data: MultiDict) -> ResponseData:
       '''Handle a form-based view.'''
       headers = {}
       if method == 'POST':
           form = ProtectedForm(form_data)
           data = {'form': form}
           if not form.validate():    # Checks the CSRF token, too!
               return data, status.BAD_REQUEST, headers

           # do something sensitive
           return data, status.CREATED, headers

       form = ProtectedForm()
       return {'form': form}, status.OK, headers


And in your template:

.. code-block:: html

   {% extends "base/base.html" %}
   {% block content %}
   <form method="POST" action="{{ url_for('ui.some_view') }}">
     {{ form.csrf_token }}
     {{ form.something_sensitive }}
     <input type="submit" class="button is-link is-small"></input>
   </form>
   {% endblock content %}

"""

import hmac
from typing import Dict, Tuple, Any
from datetime import datetime, timedelta
import dateutil.parser

from flask import request
from wtforms import Form, Field, ValidationError
from wtforms.csrf.core import CSRF
from arxiv.base.globals import get_application_config


class SessionCSRF(CSRF):
    """Session-based CSRF protection."""

    def setup_form(self, form: 'CSRFForm') -> Any:
        """Grab the CSRF context and secret from the form."""
        self.csrf_context = form.meta.csrf_context
        self.csrf_secret = form.meta.csrf_secret
        return super(SessionCSRF, self).setup_form(form)

    @staticmethod
    def _hash(secret: str, nonce: str, ip_address: str, expires: str) -> str:
        ctx = f"{nonce}{ip_address}{expires}".encode('utf-8')
        csrf_hmac = hmac.new(secret.encode('utf-8'), ctx, digestmod='sha256')
        return csrf_hmac.hexdigest()

    @staticmethod
    def _new_expiry() -> str:
        return (datetime.now() + timedelta(seconds=300)).isoformat()

    @staticmethod
    def _join(digest: str, expires: str) -> str:
        return f"{digest}::{expires}"

    @staticmethod
    def _split(csrf_token: str) -> Tuple[str, str]:
        digest, expires = csrf_token.split('::', 1)
        return digest, expires

    def generate_csrf_token(self, field: 'CSRFForm') -> str:
        """Generate a new CSRF token using the CSRF secret and context."""
        expires = self._new_expiry()
        nonce = self.csrf_context['nonce']
        ip_address = self.csrf_context['ip_address']
        digest = self._hash(self.csrf_secret, nonce, ip_address, expires)
        return self._join(digest, expires)

    def validate_csrf_token(self, form: 'CSRFForm', field: Field) -> None:
        """Validate the CSRF token passed with form data."""
        digest, expires = self._split(field.data)
        nonce = self.csrf_context['nonce']
        ip_address = self.csrf_context['ip_address']
        expected = self._hash(self.csrf_secret, nonce, ip_address, expires)
        if dateutil.parser.parse(expires) <= datetime.now():
            raise ValidationError('CSRF token has expired')
        if not hmac.compare_digest(expected, digest):
            raise ValidationError('CSRF token is invalid')


class CSRFForm(Form):
    """Base form with support for CSRF protection."""

    class Meta:
        """Set CSRF configuration."""

        csrf = True
        csrf_field_name = "csrf_token"
        csrf_class = SessionCSRF    # Set the CSRF implementation

        @property
        def csrf_secret(self) -> str:
            """The CSRF secret from the current application configuration."""
            config = get_application_config()
            try:
                secret: str = config['CSRF_SECRET']
            except KeyError as e:
                raise RuntimeError('Parameter CSRF_SECRET must be set') from e
            return secret

        @property
        def csrf_context(self) -> Dict[str, str]:
            """Session information used to generate a CSRF token."""
            if not request or not request.session:
                raise RuntimeError('Missing active user session')

            # Sessions provided by arxiv.auth should have a nonce that was
            # generated when the session was created. Legacy sessions, however,
            # do not support this. So we'll fall back to using the session ID
            # instead.
            nonce = getattr(request.session, 'nonce',
                            request.session.session_id)
            return {
                'ip_address': request.remote_addr,
                'nonce': nonce
            }



"""
Helpers for working with URLs in arXiv Flask applications.

This module provides :func:`external_url_handler`, which is attached to Flask
application instances by :class:`arxiv.base.Base`. This leverage's Flask's
built-in hook for :class:`flask.BuildError` exception handlers, described in
the :func:`flask.url_for` documentation.

To set a URL for an external (i.e. not provided by thie application) endpoint,
define a variable called ``URLS`` in your application configuration. ``URLS``
should be a list of three-tuples, of the form (``endpoint (str)``,
``path (str)``, ``domain (str)``). Paths should use the `Werkzeug rule format
<http://werkzeug.pocoo.org/docs/0.14/routing/#rule-format>`_. For example:

.. code-block:: python

   URLS = [
       ("pdf", "/pdf/<arxiv:paper_id>", "arxiv.org"),
       ("twitter", "/arxiv", "twitter.com"),
       ("blog", "/arxiv", "blogs.cornell.edu")
   ]

You can load these urls using :func:`flask.url_for`. For example,
``url_for("pdf", paper_id="1601.00123")`` should return
``https://arxiv.org/pdf/1601.00123``.

Using environment variables
---------------------------
An alternative approach is to set the endpoint using an environment variable.
Setting ``ARXIV_FOO_URL`` will define a rule for endpoint ``foo`` using the
domain and path parsed from the variable's value. For example:

.. code-block:: python

   os.environ['ARXIV_BLOG_URL'] = 'http://blogs.cornell.edu/arxiv'


Caveats
-------
URLs can be defined in :mod:`arxiv.base.config`, in environment variables,
and in the configuration of a :class:`flask.Flask` application that uses
:class:`arxiv.base.Base`. Preference is given to each of those respective
sources with increasing priority.

This will build URLs with the ``https`` scheme by default. To use ``http``,
set ``EXTERNAL_URL_SCHEME = 'http'`` in your configuration.
"""

import sys
from typing import Dict, Any
from urllib.parse import parse_qs
from werkzeug.urls import url_encode, url_parse, url_unparse
from werkzeug.routing import Map, Rule, BuildError
from flask import current_app

from arxiv.base.exceptions import ConfigurationError
from arxiv.base.converter import ArXivConverter
from arxiv.base import logging

from .clickthrough import clickthrough_url
from .links import urlize, urlizer, url_for_doi

logger = logging.getLogger(__name__)


# The module arxiv.base.config needs to be able to load its values from
# environment variables, some of which are set by SetEnv directives in Apache.
# Those variables are not set until application execution begins, which means
# that if arxiv.base.config is imported beforehand its values will not be
# correct.
def _get_base_config() -> Any:
    from arxiv.base import config
    return config


def get_url_map() -> Map:
    """Build a :class:`werkzeug.routing.Map` from configured URLs."""
    config = _get_base_config()

    # Get the base URLs (configured in this package).
    configured_urls = {url[0]: url for url in config.URLS}
    # Privilege ARXIV_URLs set on the application config.
    current_urls = current_app.config.get('URLS', [])
    if current_urls:
        configured_urls.update({url[0]: url for url in current_urls})

    url_map = Map([
        Rule(pattern, endpoint=name, host=host, build_only=True)
        for name, pattern, host in configured_urls.values()
    ], converters={'arxiv': ArXivConverter}, host_matching=True)
    return url_map


def external_url_for(endpoint: str, **values: Any) -> str:
    """
    Like :flask`.url_for`, but builds external URLs based on the config.

    This works by loading the configuration variable ``URLS`` from
    :mod:`arxiv.base.config` and from the application on which the blueprint
    has been registered, and registering the URL patterns described therein.
    Preference is given to URLs defined on the current application. An attempt
    is made to avoid adding URL rules for which identical patterns have already
    been registered.
    """
    values.pop('_external', None)
    url_map = get_url_map()
    scheme = current_app.config.get('EXTERNAL_URL_SCHEME', 'https')
    host = current_app.config.get('BASE_SERVER', 'arxiv.org')
    adapter = url_map.bind(host, url_scheme=scheme)
    url: str = adapter.build(endpoint, values=values, force_external=True)
    return url


def external_url_handler(err: BuildError, endpoint: str, values: Dict) -> str:
    """
    Attempt to handle failed URL building with :func:`external_url_for`.

    This gets attached to a Flask application via the
    :meth:`flask.Flask.url_build_error_handlers` hook.
    """
    try:
        url = external_url_for(endpoint, **values)
    except BuildError as e:
        # Re-raise the original BuildError, in context of original traceback.
        exc_type, exc_value, tb = sys.exc_info()
        if exc_value is err:
            raise exc_type(exc_value).with_traceback(tb)    # type: ignore
        else:
            raise err
    return url


def canonical_url(id: str, version: int = 0) -> str:
    """
    Generate the canonical URL for an arXiv identifier.

    This can be done from just the ID because the category is only needed if it
    is in the ID. id can be just the id or idv or cat/id or cat/idv.
    """
    # TODO: This should be better.
    # There should probably be something like INTERNAL_URL_SCHEMA
    # Also, /abs should probably be specified somewhere else
    # like arxiv.base.canonical
    scheme = current_app.config.get('EXTERNAL_URL_SCHEME', 'https')
    host = current_app.config.get('MAIN_SERVER', 'arxiv.org')
    if version:
        return f'{scheme}://{host}/abs/{id}v{version}'
    return f'{scheme}://{host}/abs/{id}'



"""URL conversion for paths containing arXiv IDs."""

import re

from arxiv import identifier
from werkzeug.routing import BaseConverter, ValidationError


class ArXivConverter(BaseConverter):
    """Route converter for arXiv IDs."""

    def to_python(self, value: str) -> str:
        """Parse URL path part to Python rep (str)."""
        try:
            return identifier.parse_arxiv_id(value)
        except ValueError as e:
            raise ValidationError('Not a valid arXiv ID') from e

    def to_url(self, value: str) -> str:
        """Cast Python rep (str) to URL path part."""
        return value



"""Application factory for Base app (development and testing only)."""

import logging
import os
from flask import Flask
from flask_s3 import FlaskS3

from arxiv.base import Base, routes

from . import config

s3 = FlaskS3()


def create_web_app() -> Flask:
    """Initialize and configure the base application."""
    app = Flask('base_test')
    # .config is an instance of a dict subclass with some methods.
    app.config.from_object(config)  # type: ignore

    Base(app)    # Gives us access to the base UI templates and resources.
    app.register_blueprint(routes.blueprint)

    s3.init_app(app)
    return app
